/*******************************************************************************
 Color Space conversions
 
 IMPLEMENTATION 

 Programmer: Dima V. Fedorov <mailto:dima@dimin.net> <http://www.dimin.net/>

 History:
   05/25/2005 14:09 - First creation

 Ver : 1
*******************************************************************************/

/* Convert between RGB and CIE-Lab color spaces
 * Uses ITU-R recommendation BT.709 with D65 as reference white.
 * Yossi Rubner 2/24/98
 *
 * Mark Ruzon 3/18/99 -- Added thresholds to truncate parts of the space
 * where changing a value has no visible effect on the color.
 */

#ifndef DIM_COLORS_H
#define DIM_COLORS_H 

#include "trims.h"
#include "round.h"

#define DIM_HSL_RANGE 255.0
#define DIM_LAB_RANGE 255.0

#define BLACK 20.0
#define YELLOW 70.0

#ifndef max
#define max(a, b)  (((a) > (b)) ? (a) : (b)) 
#endif

#ifndef min
#define min(a, b)  (((a) < (b)) ? (a) : (b))  
#endif


//----------------------------------------------------------------------------
// RGB <-> YUV
//----------------------------------------------------------------------------

inline void RGBtoYUV(const unsigned char R, const unsigned char G, const unsigned char B, 
                     unsigned char &Y, unsigned char &U, unsigned char &V)
{
  Y = trimToUChar(iround( 0.257*R + 0.504*G + 0.098*B + 16.0));
  V = trimToUChar(iround(  0.439*R - 0.368*G - 0.071*B + 128.0)); // Cr Chrominance Red
  U = trimToUChar(iround(  -0.148*R - 0.291*G + 0.439*B + 128.0)); // Cb Chrominance Blue
}

inline void YUVtoRGB(const unsigned char Y, const unsigned char U, const unsigned char V,
                     unsigned char &R, unsigned char &G, unsigned char &B )
{
  R = trimToUChar(iround( 1.164*(Y - 16.0) + 1.596*(V - 128.0) ));
  G = trimToUChar(iround( 1.164*(Y - 16.0) - 0.813*(V - 128.0) - 0.391*(U - 128.0) ));
  B = trimToUChar(iround( 1.164*(Y - 16.0) + 2.018*(U - 128.0) ));
}

//----------------------------------------------------------------------------
// RGB <-> HSL
//----------------------------------------------------------------------------

inline void RGBtoHSL(const unsigned char Rb, const unsigned char Gb, const unsigned char Bb, 
                     unsigned char &Hb, unsigned char &Sb, unsigned char &Lb)
{
  double R, G, B, D, Cmax, Cmin;
  double H, S, L;

  R = Rb / 255.0;
  G = Gb / 255.0;
  B = Bb / 255.0;

  Cmax = max (R, max (G, B));
  Cmin = min (R, min (G, B));

  // calculate luminosity
  L = (Cmax + Cmin) / 2.0;

  if (Cmax == Cmin)  // it's grey
  {
    H = 0.0; // it's actually undefined
    S = 0.0;
  } 
  else 
  {
    D = Cmax - Cmin;

    // calculate Saturation
    if (L < 0.5)
      S = D / (Cmax + Cmin);
    else
      S = D / (2.0 - Cmax - Cmin);

    // calculate Hue
    if (R == Cmax)
      H = (G - B) / D;
    else
      if (G == Cmax)
        H = 2.0 + (B - R) / D;
      else
        H = 4.0 + (R - G) / D;

    H = H / 6.0;
    if (H < 0.0) H = H + 1.0;
  }

  Hb = iround (H * (DIM_HSL_RANGE-1.0));
  Sb = iround (S * DIM_HSL_RANGE);
  Lb = iround (L * DIM_HSL_RANGE);
}


inline unsigned char hueToColourValue (double Hue, double M1, double M2)
{
  double  V;

  if (Hue < 0.0)
    Hue = Hue + 1.0;
  else
  if (Hue > 1.0)
    Hue = Hue - 1.0;

  
  if (6.0*Hue < 1.0)
    V = M1 + (M2 - M1) * Hue * 6.0;
  else
    if (2.0*Hue < 1.0)
      V = M2;
    else
      if (3.0*Hue < 2.0)
        V = M1 + (M2 - M1) * (2.0/3.0 - Hue) * 6.0;
      else
        V = M1;

  return iround (255.0 * V);
}

inline void HSLtoRGB(const unsigned char Hb, const unsigned char Sb, const unsigned char Lb,
                     unsigned char &R, unsigned char &G, unsigned char &B )
{
  double H, S, L;
  double M1, M2;

  H = Hb / (DIM_HSL_RANGE-1);
  S = Sb / DIM_HSL_RANGE;
  L = Lb / DIM_HSL_RANGE;

  if (S == 0.0)
  {
    R = iround (255.0 * L);
    G = R;
    B = R;
  }
  else
  {
    if (L <= 0.5)
      M2 = L * (1.0 + S);
    else
      M2 = L + S - L * S;
    M1 = 2.0 * L - M2;

    R = hueToColourValue (H + 1.0/3.0, M1, M2);
    G = hueToColourValue (H, M1, M2 );
    B = hueToColourValue (H - 1.0/3.0, M1, M2 );
  }
}

//----------------------------------------------------------------------------
// RGB <-> XYZ ITU D65
//----------------------------------------------------------------------------
/*    
RGB -> CIE XYZitu (D65)			| CIE XYZitu (D65) -> RGB
X = 0.431*Red+0.342*Green+0.178*Blue	| Red   =  3.063*X-1.393*Y-0.476*Z
Y = 0.222*Red+0.707*Green+0.071*Blue	| Green = -0.969*X+1.876*Y+0.042*Z
Z = 0.020*Red+0.130*Green+0.939*Blue	| Blue  =  0.068*X-0.229*Y+1.069*Z
*/



//----------------------------------------------------------------------------
// RGB <-> La*b*
//----------------------------------------------------------------------------
/*
  L [0-100]
  a [-0.5 0.5]
  b [-0.5 0.5]

  */

inline void RGB2Lab(float R, float G, float B, float *L, float *a, float *b)
{
  float X, Y, Z, fX, fY, fZ;

  X = 0.412453*R + 0.357580*G + 0.180423*B;
  Y = 0.212671*R + 0.715160*G + 0.072169*B;
  Z = 0.019334*R + 0.119193*G + 0.950227*B;

  X /= (255.0F * 0.950456F);
  Y /=  255;
  Z /= (255.0F * 1.088754F);

  if (Y > 0.008856)
    {
      fY = pow(Y, 1.0/3.0);
      *L = 116.0*fY - 16.0;
    }
  else
    {
      fY = 7.787*Y + 16.0/116.0;
      *L = 903.3*Y;
    }

  if (X > 0.008856)
      fX = pow(X, 1.0/3.0);
  else
      fX = 7.787*X + 16.0/116.0;

  if (Z > 0.008856)
      fZ = pow(Z, 1.0/3.0);
  else
      fZ = 7.787*Z + 16.0/116.0;

  *a = 500.0*(fX - fY);
  *b = 200.0*(fY - fZ);

  if (*L < BLACK) {
    *a *= exp((*L - BLACK) / (BLACK / 4));
    *b *= exp((*L - BLACK) / (BLACK / 4));
    *L = BLACK;
  }
  if (*b > YELLOW)
    *b = YELLOW;
}


inline void Lab2RGB(float L, float a, float b, float *R, float *G, float *B)
{
  float X, Y, Z, fX, fY, fZ;
  float RR, GG, BB;

  fY = pow((L + 16.0) / 116.0, 3.0);
  if (fY < 0.008856)
      fY = L / 903.3;
  Y = fY;

  if (fY > 0.008856)
    fY = pow(fY, 1.0/3.0);
  else
    fY = 7.787 * fY + 16.0/116.0;

  fX = a / 500.0 + fY;      
  if (fX > 0.206893)
      X = pow(fX, 3.0);
  else
      X = (fX - 16.0/116.0) / 7.787;
 
  fZ = fY - b /200.0;      
  if (fZ > 0.206893)
      Z = pow(fZ, 3.0);
  else
      Z = (fZ - 16.0/116.0) / 7.787;

  X *= (0.950456F * 255.0F);
  Y *=             255;
  Z *= (1.088754F * 255.0F);

  RR =  3.240479*X - 1.537150*Y - 0.498535*Z;
  GG = -0.969256*X + 1.875992*Y + 0.041556*Z;
  BB =  0.055648*X - 0.204043*Y + 1.057311*Z;

  *R = RR;
  *G = GG;
  *B = BB;
}


inline void RGBtoLAB(const unsigned char Rb, const unsigned char Gb, const unsigned char Bb,
                     unsigned char &lb, unsigned char &ab, unsigned char &bb )
{
  float l, a, b;

  RGB2Lab(Rb, Gb, Bb, &l, &a, &b);
  lb = trimToUChar(iround(l * 2.55));
  ab = trimToUChar(iround(a * 12.75 + 128));
  bb = trimToUChar(iround(b * 12.75 + 128));
}


inline void LABtoRGB(const unsigned char lb, const unsigned char ab, const unsigned char bb,
                     unsigned char &Rb, unsigned char &Gb, unsigned char &Bb )
{
  float l, a, b;
  float R, G, B;

  l = (lb / 2.55);
  a = (ab-128 / 12.75);
  b = (bb-128 / 12.75);

  Lab2RGB(l, a, b, &R, &G, &B);

  Rb = trimToUChar(iround( R ));
  Gb = trimToUChar(iround( G ));
  Bb = trimToUChar(iround( B ));
}

#endif //DIM_COLORS_H 


